/*
 * Sixth 3D engine. Copyright ©2012-2017, Svjatoslav Agejenko, svjatoslav@svjatoslav.eu
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of version 3 of the GNU Lesser General Public License
 * or later as published by the Free Software Foundation.
 *
 */

package eu.svjatoslav.sixth.e3d.gui;

import eu.svjatoslav.sixth.e3d.geometry.Point3D;

public class Avatar implements ViewUpdateListener {

    public static final double SPEED_LIMIT = 30;
    /**
     * Just in case we want to adjust global speed for some reason.
     */
    private static final double SPEED_MULTIPLIER = .02d;
    /**
     * Avatar movement speed, relative to avatar itself. When avatar coordinates
     * are updated within the world, avatar orientation relative to the world is
     * taken into account.
     */
    private final Point3D movementDirection = new Point3D();
    public double avatarAcceleration = 0.1;
    /**
     * Avatar location within the 3D world.
     */
    private Point3D location = new Point3D();

    /**
     * Avatar orientation on the X-Z plane. It changes when turning left or
     * right.
     */
    private double orientationXZ;

    /**
     * Avatar orientation on the Y-Z plane. It changes when looking up or down.
     */
    private double orientationYZ;

    public Avatar() {
    }

    public Avatar(final Avatar sourceView) {
        setLocation(new Point3D(sourceView.location));
        setAngleXZ(sourceView.getAngleXZ());
        setAngleYZ(sourceView.getAngleYZ());
    }

    public Avatar(final Point3D location) {
        setLocation(location);
    }

    public Avatar(final Point3D location, final float angleXZ,
                  final float angleYZ) {
        setLocation(location);
        setAngleXZ(angleXZ);
        setAngleYZ(angleYZ);
    }

    @Override
    public boolean beforeViewUpdate(final ViewContext viewContext,
                                    final int millisecondsSinceLastFrame) {

        final Point3D locationBeforeUpdate = new Point3D(location);
        updateLocation(millisecondsSinceLastFrame);

        final double distanceMoved = location
                .getDistanceTo(locationBeforeUpdate);

        return distanceMoved > 0.03;

    }

    public void enforceSpeedLimit() {
        final double currentSpeed = movementDirection
                .getDistanceTo(Point3D.ZERO);

        if (currentSpeed <= SPEED_LIMIT)
            return;

        movementDirection.scaleDown(currentSpeed / SPEED_LIMIT);
    }

    public double getAngleXZ() {
        return orientationXZ;
    }

    public void setAngleXZ(final double angleXZ) {
        orientationXZ = angleXZ;
    }

    public double getAngleYZ() {
        return orientationYZ;
    }

    public void setAngleYZ(final double angleYZ) {
        orientationYZ = angleYZ;
    }

    public Point3D getLocation() {
        return location;
    }

    public void setLocation(final Point3D location) {
        this.location = location;
    }

    public Point3D getMovementDirection() {
        return movementDirection;
    }

    public double getMovementSpeed() {
        return movementDirection.getDistanceTo(Point3D.ZERO);
    }

    /**
     * Update camera location based on current speed
     */
    public void updateLocation(final int millisecondsPassedSinceLastFrame) {

        // translate user coordinates based on avatar movement speed, and avatar
        // orientation in the world
        {
            location.x -= (float) Math.sin(getAngleXZ())
                    * getMovementDirection().z * SPEED_MULTIPLIER
                    * millisecondsPassedSinceLastFrame;
            location.z += (float) Math.cos(getAngleXZ())
                    * getMovementDirection().z * SPEED_MULTIPLIER
                    * millisecondsPassedSinceLastFrame;

            location.x += (float) Math.cos(getAngleXZ())
                    * getMovementDirection().x * SPEED_MULTIPLIER
                    * millisecondsPassedSinceLastFrame;
            location.z += (float) Math.sin(getAngleXZ())
                    * getMovementDirection().x * SPEED_MULTIPLIER
                    * millisecondsPassedSinceLastFrame;

            location.y += getMovementDirection().y * SPEED_MULTIPLIER
                    * millisecondsPassedSinceLastFrame;
        }

        final double millisecondFriction = 1.005;
        // apply friction to progressively slow movement
        for (int i = 0; i < millisecondsPassedSinceLastFrame; i++) {
            getMovementDirection().x = getMovementDirection().x
                    / millisecondFriction;
            getMovementDirection().y = getMovementDirection().y
                    / millisecondFriction;
            getMovementDirection().z = getMovementDirection().z
                    / millisecondFriction;
        }
    }
}
