/*
 * Sixth 3D engine. Copyright ©2012-2017, Svjatoslav Agejenko, svjatoslav@svjatoslav.eu
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of version 3 of the GNU Lesser General Public License
 * or later as published by the Free Software Foundation.
 *
 */

package eu.svjatoslav.sixth.e3d.gui.humaninput;

import eu.svjatoslav.sixth.e3d.geometry.Point2D;
import eu.svjatoslav.sixth.e3d.gui.Avatar;
import eu.svjatoslav.sixth.e3d.gui.View;
import eu.svjatoslav.sixth.e3d.gui.ViewContext;
import eu.svjatoslav.sixth.e3d.gui.ViewUpdateListener;

import javax.swing.*;
import java.awt.event.*;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

public class UserInputTracker
        implements MouseMotionListener, KeyListener, MouseListener, MouseWheelListener, ViewUpdateListener {

    /**
     * <pre>
     * Key is keyboard key code.
     * Value is system milliseconds when key was pressed.
     *
     * So by reading the map one can determine currently pressed keys as well as duration.
     * </pre>
     */
    private final Map<Integer, Long> pressedKeysToPressedTimeMap = new HashMap<>();
    private final List<MouseClick> detectedMouseClicks = new ArrayList<>();
    private final List<KeyEvent> detectedKeyEvents = new ArrayList<>();
    public int wheelMovedDirection = 0;
    Point2D mouseDraggedDirection = new Point2D();
    Point2D oldMouseCoordinatesWhenDragging;
    ViewContext viewContext;
    Point2D currentMouseLocation;
    boolean mouseMoved;
    private boolean mouseWithinWindow = false;

    public UserInputTracker(final ViewContext viewContext) {
        this.viewContext = viewContext;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public boolean beforeViewUpdate(final ViewContext viewContext, final int millisecondsSinceLastFrame) {

        boolean viewUpdateNeeded = handleDetectedMouseClicks(viewContext.getView());

        viewUpdateNeeded |= handleDetectedKeyEvents();

        viewContext.getKeyboardFocusTracker().getCurrentFocusOwner().beforeViewUpdate(viewContext,
                millisecondsSinceLastFrame);
        viewUpdateNeeded |= trackMouse();

        return viewUpdateNeeded;
    }

    public void bind(final JPanel panel) {
        panel.addMouseMotionListener(this);

        panel.addKeyListener(this);

        panel.addMouseListener(this);

        panel.addMouseWheelListener(this);
    }

    public boolean handleDetectedKeyEvents() {
        boolean keyEventsHandled = false;

        final UserInputHandler currentFocusOwner = viewContext.getKeyboardFocusTracker().getCurrentFocusOwner();

        synchronized (detectedKeyEvents) {
            if (currentFocusOwner == null) {
                detectedKeyEvents.clear();
                return false;
            }

            while (!detectedKeyEvents.isEmpty()) {
                final KeyEvent keyEvent = detectedKeyEvents.remove(0);

                switch (keyEvent.getID()) {
                    case KeyEvent.KEY_PRESSED:
                        currentFocusOwner.keyPressed(keyEvent, viewContext);
                        keyEventsHandled = true;
                        break;

                    case KeyEvent.KEY_RELEASED:
                        currentFocusOwner.keyReleased(keyEvent, viewContext);
                        keyEventsHandled = true;
                        break;
                }
            }
        }

        return keyEventsHandled;
    }

    /**
     * Returns <code>true</code> if mouse events are detected and view needs to
     * be repainted.
     */
    public synchronized boolean handleDetectedMouseClicks(final View view) {
        if (detectedMouseClicks.isEmpty()) {

            if (currentMouseLocation != null)
                view.getRenderingContext().mouseClick = new MouseClick(currentMouseLocation, 0);

            if (mouseMoved) {
                mouseMoved = false;
                return true;
            } else
                return false;
        }

        view.getRenderingContext().mouseClick = detectedMouseClicks.remove(0);

        return true;
    }

    public boolean isKeyPressed(final int keyCode) {
        return pressedKeysToPressedTimeMap.containsKey(keyCode);
    }

    @Override
    public void keyPressed(final KeyEvent evt) {
        synchronized (detectedKeyEvents) {
            pressedKeysToPressedTimeMap.put(evt.getKeyCode(), System.currentTimeMillis());
            detectedKeyEvents.add(evt);
            viewContext.getView().repaintDuringNextViewUpdate();
        }
    }

    @Override
    public void keyReleased(final KeyEvent evt) {
        synchronized (detectedKeyEvents) {
            pressedKeysToPressedTimeMap.remove(evt.getKeyCode());
            detectedKeyEvents.add(evt);
            viewContext.getView().repaintDuringNextViewUpdate();
        }
    }

    @Override
    public void keyTyped(final KeyEvent e) {
    }

    @Override
    public void mouseClicked(final MouseEvent e) {
        synchronized (detectedMouseClicks) {
            detectedMouseClicks.add(new MouseClick(e.getX(), e.getY(), e.getButton()));
        }
    }

    @Override
    public void mouseDragged(final java.awt.event.MouseEvent evt) {
        final Point2D mouseLocation = new Point2D(evt.getX(), evt.getY());

        if (oldMouseCoordinatesWhenDragging == null) {
            oldMouseCoordinatesWhenDragging = mouseLocation;
            return;
        }

        mouseDraggedDirection.add(mouseLocation.clone().subtract(oldMouseCoordinatesWhenDragging));

        oldMouseCoordinatesWhenDragging = mouseLocation;
    }

    @Override
    public void mouseEntered(final MouseEvent e) {
        mouseWithinWindow = true;
    }

    @Override
    public synchronized void mouseExited(final MouseEvent e) {
        mouseWithinWindow = false;
        currentMouseLocation = null;
    }

    @Override
    public synchronized void mouseMoved(final MouseEvent e) {
        currentMouseLocation = new Point2D(e.getX(), e.getY());
        mouseMoved = true;
    }

    @Override
    public void mousePressed(final MouseEvent e) {
    }

    @Override
    public void mouseReleased(final java.awt.event.MouseEvent evt) {
        oldMouseCoordinatesWhenDragging = null;
    }

    @Override
    public void mouseWheelMoved(final java.awt.event.MouseWheelEvent evt) {
        wheelMovedDirection += evt.getWheelRotation();
    }

    /**
     * Interpret mouse movement
     */
    public boolean trackMouse() {
        final Avatar avatar = viewContext.getAvatar();

        // track mouse dragging

        // TODO: need to detect whether user moved mouse or touch screen

        // for mouse
        avatar.setAngleXZ(avatar.getAngleXZ() - ((float) mouseDraggedDirection.x / 50));
        avatar.setAngleYZ(avatar.getAngleYZ() - ((float) mouseDraggedDirection.y / 50));

        // for touch screen
        // avatar.setAngleXZ(avatar.getAngleXZ() + ((float)
        // mouseDraggedDirection.x / 50));
        // avatar.setAngleYZ(avatar.getAngleYZ() + ((float)
        // mouseDraggedDirection.y / 50));

        // track mouse wheel movements
        final double actualAcceleration = 50 * avatar.avatarAcceleration * (1 + (avatar.getMovementSpeed() / 10));

        avatar.getMovementDirection().y += (wheelMovedDirection * actualAcceleration);
        avatar.enforceSpeedLimit();

        // check if view shall be repainted
        boolean repaintNeeded;
        repaintNeeded = (mouseDraggedDirection.x != 0) || (mouseDraggedDirection.y != 0) || (wheelMovedDirection != 0);

        // reset movement counters
        wheelMovedDirection = 0;
        mouseDraggedDirection.zero();

        return repaintNeeded;
    }

}
