/*
 * Sixth 3D engine. Copyright ©2012-2017, Svjatoslav Agejenko, svjatoslav@svjatoslav.eu
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of version 3 of the GNU Lesser General Public License
 * or later as published by the Free Software Foundation.
 *
 */

package eu.svjatoslav.sixth.e3d.renderer.raster.shapes.basic.texturedpolygon;

import eu.svjatoslav.sixth.e3d.geometry.Point2D;
import eu.svjatoslav.sixth.e3d.geometry.Point3D;
import eu.svjatoslav.sixth.e3d.geometry.Polygon;
import eu.svjatoslav.sixth.e3d.gui.RenderingContext;
import eu.svjatoslav.sixth.e3d.renderer.raster.shapes.AbstractCoordinateShape;
import eu.svjatoslav.sixth.e3d.renderer.raster.slicer.PolygonCoordinate;
import eu.svjatoslav.sixth.e3d.renderer.raster.texture.Texture;
import eu.svjatoslav.sixth.e3d.renderer.raster.texture.TextureBitmap;

import java.awt.*;

public class TexturedPolygon extends AbstractCoordinateShape {

    public final Texture texture;
    final PolygonBorderInterpolator[] is = new PolygonBorderInterpolator[]{
            new PolygonBorderInterpolator(), new PolygonBorderInterpolator(),
            new PolygonBorderInterpolator()};
    /**
     * Polygon texture coordinates.
     */
    public Point2D texturePoint1, texturePoint2, texturePoint3;
    private boolean showBorders = false;
    private double totalTextureDistance = -1;

    public TexturedPolygon(final Point3D p1, final Point3D p2,
                           final Point3D p3, final Point2D tp1, final Point2D tp2,
                           final Point2D tp3, final Texture texture) {

        super(p1, p2, p3);

        texturePoint1 = tp1;
        texturePoint2 = tp2;
        texturePoint3 = tp3;

        this.texture = texture;
    }

    public TexturedPolygon(final PolygonCoordinate pc1,
                           final PolygonCoordinate pc2, final PolygonCoordinate pc3,
                           final Texture texture) {

        this(pc1.space, pc2.space, pc3.space, pc1.texture, pc2.texture,
                pc3.texture, texture);
    }

    private void computeTotalTextureDistance() {
        // compute total texture distance
        totalTextureDistance = texturePoint1.getDistanceTo(texturePoint2);
        totalTextureDistance += texturePoint1.getDistanceTo(texturePoint3);
        totalTextureDistance += texturePoint2.getDistanceTo(texturePoint3);
    }

    private void drawHorizontalLine(final PolygonBorderInterpolator line1,
                                    final PolygonBorderInterpolator line2, final int y,
                                    final RenderingContext renderBuffer,
                                    final TextureBitmap textureBitmap) {

        line1.setCurrentY(y);
        line2.setCurrentY(y);

        int x1 = line1.getX();
        int x2 = line2.getX();

        final double tx2, ty2;
        final double tx1, ty1;

        if (x1 <= x2) {

            tx1 = line1.getTX() * textureBitmap.multiplicationFactor;
            ty1 = line1.getTY() * textureBitmap.multiplicationFactor;

            tx2 = line2.getTX() * textureBitmap.multiplicationFactor;
            ty2 = line2.getTY() * textureBitmap.multiplicationFactor;

        } else {
            final int tmp = x1;
            x1 = x2;
            x2 = tmp;

            tx1 = line2.getTX() * textureBitmap.multiplicationFactor;
            ty1 = line2.getTY() * textureBitmap.multiplicationFactor;

            tx2 = line1.getTX() * textureBitmap.multiplicationFactor;
            ty2 = line1.getTY() * textureBitmap.multiplicationFactor;
        }

        final double realWidth = x2 - x1;
        final double realX1 = x1;

        if (x1 < 0)
            x1 = 0;

        if (x2 >= renderBuffer.width)
            x2 = renderBuffer.width - 1;

        int renderBufferOffset = ((y * renderBuffer.width) + x1) * 4;
        final byte[] renderBufferBytes = renderBuffer.bytes;

        final double twidth = tx2 - tx1;
        final double theight = ty2 - ty1;

        for (int x = x1; x < x2; x++) {

            final double distance = x - realX1;

            final double tx = tx1 + ((twidth * distance) / realWidth);
            final double ty = ty1 + ((theight * distance) / realWidth);

            final int textureOffset = textureBitmap.getAddress((int) tx,
                    (int) ty);

            textureBitmap.drawPixel(textureOffset, renderBufferBytes,
                    renderBufferOffset);

            renderBufferOffset += 4;
        }

    }

    @Override
    public void paint(final RenderingContext renderBuffer) {

        final Point2D projectedPoint1 = coordinates[0].onScreenCoordinate;
        final Point2D projectedPoint2 = coordinates[1].onScreenCoordinate;
        final Point2D projectedPoint3 = coordinates[2].onScreenCoordinate;

        projectedPoint1.roundToInteger();
        projectedPoint2.roundToInteger();
        projectedPoint3.roundToInteger();

        if (mouseInteractionController != null)
            if (renderBuffer.mouseClick != null)
                if (Polygon.pointWithinPolygon(
                        renderBuffer.mouseClick.coordinate, projectedPoint1,
                        projectedPoint2, projectedPoint3))
                    renderBuffer.clickedItem = mouseInteractionController;

        // Show polygon boundaries (for debugging)
        if (showBorders)
            showBorders(renderBuffer);

        // find top-most point
        int yTop = (int) projectedPoint1.y;

        if (projectedPoint2.y < yTop)
            yTop = (int) projectedPoint2.y;

        if (projectedPoint3.y < yTop)
            yTop = (int) projectedPoint3.y;

        if (yTop < 0)
            yTop = 0;

        // find bottom-most point
        int yBottom = (int) projectedPoint1.y;

        if (projectedPoint2.y > yBottom)
            yBottom = (int) projectedPoint2.y;

        if (projectedPoint3.y > yBottom)
            yBottom = (int) projectedPoint3.y;

        if (yBottom >= renderBuffer.height)
            yBottom = renderBuffer.height - 1;

        // paint
        double totalVisibleDistance = projectedPoint1
                .getDistanceTo(projectedPoint2);
        totalVisibleDistance += projectedPoint1.getDistanceTo(projectedPoint3);
        totalVisibleDistance += projectedPoint2.getDistanceTo(projectedPoint3);

        if (totalTextureDistance == -1)
            computeTotalTextureDistance();
        final double scaleFactor = (totalVisibleDistance / totalTextureDistance) * 1.2d;

        final TextureBitmap zoomedBitmap = texture.getZoomedBitmap(scaleFactor);

        is[0].setPoints(projectedPoint1, projectedPoint2, texturePoint1,
                texturePoint2);
        is[1].setPoints(projectedPoint1, projectedPoint3, texturePoint1,
                texturePoint3);
        is[2].setPoints(projectedPoint2, projectedPoint3, texturePoint2,
                texturePoint3);

        java.util.Arrays.sort(is);

        for (int y = yTop; y < yBottom; y++)
            if (is[0].containsY(y)) {

                if (is[1].containsY(y))
                    drawHorizontalLine(is[0], is[1], y, renderBuffer,
                            zoomedBitmap);
                else if (is[2].containsY(y))
                    drawHorizontalLine(is[0], is[2], y, renderBuffer,
                            zoomedBitmap);
            } else if (is[1].containsY(y))
                if (is[2].containsY(y))
                    drawHorizontalLine(is[1], is[2], y, renderBuffer,
                            zoomedBitmap);

    }

    private void showBorders(final RenderingContext renderBuffer) {

        final Point2D projectedPoint1 = coordinates[0].onScreenCoordinate;
        final Point2D projectedPoint2 = coordinates[1].onScreenCoordinate;
        final Point2D projectedPoint3 = coordinates[2].onScreenCoordinate;

        renderBuffer.graphics.setColor(Color.YELLOW);
        renderBuffer.graphics.drawLine((int) projectedPoint1.x,
                (int) projectedPoint1.y, (int) projectedPoint2.x,
                (int) projectedPoint2.y);
        renderBuffer.graphics.drawLine((int) projectedPoint3.x,
                (int) projectedPoint3.y, (int) projectedPoint2.x,
                (int) projectedPoint2.y);
        renderBuffer.graphics.drawLine((int) projectedPoint1.x,
                (int) projectedPoint1.y, (int) projectedPoint3.x,
                (int) projectedPoint3.y);
    }

}
